#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <gtk/gtk.h>
#include <string.h>

#include "xfbib-bibtex-value.h"
#include "xfbib-bibtex-string.h"
#include "xfbib-bibtex-preamble.h"
#include "xfbib-bibtex.h"
#include "xfbib-string.h"
#include "xfbib-integer.h"

static gint bibtex_string_cmp(gconstpointer, gconstpointer);

struct _XfbibBibtexValue
{
	GObject parent;
	GObject *owner;
	GList *list;
};

typedef struct _XfbibBibtexValueClass
{
	GObjectClass parent;
} XfbibBibtexValueClass;

static void xfbib_bibtex_value_class_init(XfbibBibtexValueClass *klass);

static void xfbib_bibtex_value_init(XfbibBibtexValue *instance);
static void xfbib_bibtex_value_finalize(GObject *obj);

G_DEFINE_TYPE(XfbibBibtexValue, xfbib_bibtex_value, G_TYPE_OBJECT)

static void
xfbib_bibtex_value_class_init(XfbibBibtexValueClass *klass)
{
	GObjectClass *object_class = (GObjectClass *)klass;
	object_class->finalize = xfbib_bibtex_value_finalize;
}

static void
xfbib_bibtex_value_init(XfbibBibtexValue *instance)
{
	instance->list = NULL;
}

static void
xfbib_bibtex_value_finalize(GObject *obj)
{
	XfbibBibtexValue *value;
	g_return_if_fail(XFBIB_IS_BIBTEX_VALUE(obj));
	
	value = XFBIB_BIBTEX_VALUE(obj);
	g_list_foreach (value->list, (GFunc) g_object_unref, NULL);
	g_list_free (value->list);
	value->list = NULL;
	G_OBJECT_CLASS(xfbib_bibtex_value_parent_class)->finalize(obj);
}

XfbibBibtexValue *
xfbib_bibtex_value_new(GObject *obj)
{
	XfbibBibtexValue *value;
	value = g_object_new(XFBIB_TYPE_BIBTEX_VALUE, NULL);
	value->owner = obj;
	return value;
}

GObject *
xfbib_bibtex_value_get_owner(XfbibBibtexValue *value)
{
	g_return_val_if_fail(XFBIB_IS_BIBTEX_VALUE(value), NULL);
	return value->owner;
}

gboolean
xfbib_bibtex_value_parse(XfbibBibtexValue *value, const gchar *str)
{
	gint i, j, len, braces;
	gchar **split = NULL;
	GObject *obj;
	XfbibBibtex *elements;

	if (XFBIB_IS_BIBTEX_STRING(value->owner)) {
		elements = XFBIB_BIBTEX(xfbib_bibtex_string_get_owner(XFBIB_BIBTEX_STRING(value->owner)));
	} else if (XFBIB_IS_BIBTEX_FIELD(value->owner)) {
		elements = XFBIB_BIBTEX(xfbib_bibtex_entry_get_owner(
				XFBIB_BIBTEX_ENTRY(xfbib_bibtex_field_get_owner(XFBIB_BIBTEX_FIELD(value->owner)))));
	} else if (XFBIB_IS_BIBTEX_PREAMBLE(value->owner)) {
		elements = XFBIB_BIBTEX(xfbib_bibtex_preamble_get_owner(XFBIB_BIBTEX_PREAMBLE(value->owner)));
	} else {
		return FALSE;
	}

	/* TODO: Free all elements of the list */
	value->list = NULL;

	split = g_strsplit(str, "#", 0);

	for (i = 0; split[i] != NULL; i++) {
		/*
		 * Check the type of the string to parse.
		 *
		 * If the string starts with " or { it is a string,
		 * if the string starts with a alphabetic character
		 * it is a variable or if it contains all numbers it
		 * is an integer.
		 */
		g_strstrip(split[i]);
		if (split[i][0] == '"' || split[i][0] == '{') {
			/* String */

			len = strlen(split[i]);

			if (split[i][0] == '"' && split[i][len - 1] != '"')
				return FALSE;
			else if (split[i][0] == '{' && split[i][len - 1] != '}')
				return FALSE;

			if (split[i][0] == '{' && strchr(split[i], '@') != NULL) {
				/* Braced string contains @ */
				g_warning(_("Braced string contains @: %s\n"), split[i]);
				return FALSE;
			}

			braces = 0;
			len = strlen(split[i]) - 1;
			for (j = 1; j < len; j++) {
				/* Check if the braces are balanced and quotes are inside braces*/
				if (split[i][j] == '{')
					braces++;
				else if (split[i][j] == '}')
					braces--;
				else if (split[i][0] == '"'  && split[i][j] == '"' && braces == 0) {
					g_warning(_("Double quotation marks not inside braces in string: %s\n"), split[i]);
					return FALSE;
				}
			}

			if (braces != 0) {
				/* Braces was not balanced */
				g_warning(_("Braces not balanced in field string: %s\n"), split[i]);
				return FALSE;
			}

			obj = G_OBJECT(xfbib_string_new(G_OBJECT(value)));
			xfbib_string_parse(XFBIB_STRING(obj), split[i]);
			value->list = g_list_prepend(value->list, obj);
		} else if(g_ascii_isalpha(split[i][0])) {
			/* Variable */
			if (strstr(split[i], "\"") != NULL) {
				g_warning(_("Variable contains double quotation marks: %s\n"), split[i]);
				return FALSE;
			}

			obj = xfbib_bibtex_find(elements, bibtex_string_cmp, split[i]);
			if(obj != NULL) {
				value->list = g_list_prepend(value->list, obj);
			}
		} else {
			len = strlen(split[i]);
			for (j = 0; j < len; j++) {
				if (!g_ascii_isdigit(split[i][j])) {
					/* Character was not a number between 0 and 9 */
					g_warning(_("Number contains a non-digit: %s\n"), split[i]);
					return FALSE;
				}
			}
			/* Integer */
			obj = G_OBJECT(xfbib_integer_new(G_OBJECT(value)));
			xfbib_integer_set(XFBIB_INTEGER(obj), split[i]);
			value->list = g_list_prepend(value->list, obj);
		}
	}

	g_strfreev(split);


	value->list = g_list_reverse(value->list);
	return TRUE;
}

GList *
xfbib_bibtex_value_get(XfbibBibtexValue *value)
{
	g_return_val_if_fail(XFBIB_IS_BIBTEX_VALUE(value), NULL);
	
	return value->list;
}


void 
xfbib_bibtex_value_set(XfbibBibtexValue *value, GList *list)
{
	g_return_if_fail(XFBIB_IS_BIBTEX_VALUE(value));

	/* Remove all values */
	g_list_foreach (value->list, (GFunc) g_object_unref, NULL);
	g_list_free (value->list);
	
	value->list = list;
}

/* 
 * Returned string must be freed
 */
gchar *
xfbib_bibtex_value_get_str(XfbibBibtexValue *value)
{
	XfbibBibtexValue *val;
	GString *str;
	GList *list;
	gchar *tmp;
	g_return_val_if_fail(XFBIB_IS_BIBTEX_VALUE(value), NULL);
	
	str = g_string_new(NULL);

	for (list = value->list; list != NULL; list = g_list_next(list)) {
		if (XFBIB_IS_STRING(list->data)) {
			/* regular string */
			g_string_append(str, xfbib_string_get(XFBIB_STRING(list->data)));
		} else if (XFBIB_IS_INTEGER(list->data)) {
			/* integer value */
			g_string_append(str, xfbib_integer_get(XFBIB_INTEGER(list->data)));
		} else if (XFBIB_IS_BIBTEX_STRING(list->data)) {
			/* variable */
			val = xfbib_bibtex_string_get_value(XFBIB_BIBTEX_STRING(list->data));
			tmp = xfbib_bibtex_value_get_str(val);
			g_string_append(str, tmp);
			g_free(tmp);
		}
	}
	return g_string_free(str, FALSE);
}

/*
 * Remove all values in the list and set the first value to str
 */
void
xfbib_bibtex_value_set_str(XfbibBibtexValue *value, const gchar *str)
{
	XfbibString *string;
	g_return_if_fail(XFBIB_IS_BIBTEX_VALUE(value));
	
	/* Remove all values */
	g_list_foreach (value->list, (GFunc) g_object_unref, NULL);
	g_list_free (value->list);
	value->list = NULL;
	
	string = xfbib_string_new(G_OBJECT(value));
	xfbib_string_set(string, str);
	value->list = g_list_append(value->list, G_OBJECT(string));	
}

static gint 
bibtex_string_cmp(gconstpointer a, gconstpointer b)
{
	g_return_val_if_fail(a != NULL, -1);
	if(!XFBIB_IS_BIBTEX_STRING(b)) {
		return 1;	
	}

	return strcmp(a, xfbib_bibtex_string_get_variable((XfbibBibtexString *)b));
}

