#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include <gtk/gtk.h>

#include "xfbib-bibtex.h"
#include "xfbib-bibtex-entry.h"
#include "xfbib-bibtex-comment.h"
#include "xfbib-bibtex-preamble.h"
#include "xfbib-bibtex-string.h"

#ifndef G_SEQUENCE
#define G_SEQUENCE GLIB_CHECK_VERSION (2, 14, 0)
#endif


#define CURLY_BRACKET 0
#define ROUND_BRACKET 1

struct _XfbibBibtex
{
	GObject parent;
	gint n_entries;
	gint n_strings;
#if G_SEQUENCE
	GSequence *seq;
#else
	GList *list;
#endif
};

typedef struct _XfbibBibtexClass
{
	GObjectClass parent;
} XfbibBibtexClass;

static void xfbib_bibtex_class_init(XfbibBibtexClass *klass);

static void xfbib_bibtex_init(XfbibBibtex *instance);
static void xfbib_bibtex_finalize(GObject *obj);

G_DEFINE_TYPE(XfbibBibtex, xfbib_bibtex, G_TYPE_OBJECT)

static void
xfbib_bibtex_class_init(XfbibBibtexClass *klass)
{
	GObjectClass *object_class = (GObjectClass *)klass;
	object_class->finalize = xfbib_bibtex_finalize;
}

static void
xfbib_bibtex_init(XfbibBibtex *instance)
{
#if G_SEQUENCE
	instance->seq = g_sequence_new(g_object_unref);
#else
	instance->list = NULL;
#endif
	instance->n_entries = 0;
	instance->n_strings = 0;

}

static void
xfbib_bibtex_finalize(GObject *obj)
{
	/* Free the Bibtex List/Sequence */
	xfbib_bibtex_remove_all(XFBIB_BIBTEX(obj));
	G_OBJECT_CLASS(xfbib_bibtex_parent_class)->finalize(obj);
}

XfbibBibtex *
xfbib_bibtex_new()
{
	XfbibBibtex *elements;
	XfbibBibtexString *string;
	gchar months[12][32] = {
			"@String{jan = \"january\"}",
			"@String{feb = \"february\"}",
			"@String{mar = \"march\"}",
			"@String{apr = \"april\"}",
			"@String{may = \"may\"}",
			"@String{jun = \"june\"}",
			"@String{jul = \"july\"}",
			"@String{aug = \"august\"}",
			"@String{sep = \"september\"}",
			"@String{oct = \"october\"}",
			"@String{nov = \"november\"}",
			"@String{dec = \"december\"}"};
	gint n;

	elements = g_object_new(XFBIB_TYPE_BIBTEX, NULL);

	/* Add default bibtex strings */
	for(n = 0; n < 12; n++) {
		string = xfbib_bibtex_string_new(G_OBJECT(elements));
		if (xfbib_bibtex_string_parse(string, months[n])) {
			xfbib_bibtex_string_set_internal(string, 1);
			elements->n_strings++;
#if G_SEQUENCE
			g_sequence_append(elements->seq, string);
#else
			elements->list = g_list_append(elements->list, string);
#endif
		}
	}

	return elements;
}


static gboolean
parse_balanced_brackets(GString *buf, const gchar *str, gint *i)
{
	gint brackets;
	gchar bracket_type;

	while (str[(*i)] != '{' && str[(*i)] != '(') {
		buf = g_string_append_c(buf, str[(*i)]);
		(*i)++;
	}

	if (str[(*i)] == '{') {
		bracket_type = CURLY_BRACKET;
	} else if (str[(*i)] == '(') {
		bracket_type = ROUND_BRACKET;
	} else {
		g_warning(_("Unknown bracket type: %c\n"), str[(*i)]);
		return FALSE;
	}

	buf = g_string_append_c(buf, str[(*i)]);
	(*i)++;

	for (brackets = 1; brackets > 0; (*i)++) {
		if (bracket_type == CURLY_BRACKET) {
			if (str[(*i)] == '{')
				brackets++;
			else if(str[(*i)] == '}')
				brackets--;
		} else if (bracket_type == ROUND_BRACKET) {
			if (str[(*i)] == '(')
				brackets++;
			else if(str[(*i)] == ')')
				brackets--;
		}
		buf = g_string_append_c(buf, str[(*i)]);
	}
	return TRUE;
}


gboolean
xfbib_bibtex_parse(XfbibBibtex *elements, const gchar *str)
{
	gint i, len;
	GRegex *regex;
	GObject *obj;
	GString *buf;

	buf = g_string_new("");

	/* Compile the regex for the entry types */
	regex = g_regex_new("^@(article|book|booklet|conference|inbook|incollection|inproceedings|manual|"
			"mastersthesis|misc|phdthesis|proceedings|techreport|unpublished)({|\\s|[(])",
			G_REGEX_CASELESS, 0, NULL);
	
	len = strlen(str);
	for (i = 0; i < len; i++) {
		if (str[i] == '@') {
			if (buf->len > 0) {
				/* The comment ended put it in the list */
				obj = G_OBJECT(xfbib_bibtex_comment_new(G_OBJECT(elements), buf->str));
#if G_SEQUENCE
				g_sequence_append(elements->seq, obj);
#else
				elements->list = g_list_append(elements->list, obj);
#endif
				buf = g_string_erase(buf, 0, -1);
			}
			if (g_ascii_strncasecmp(&str[i], "@comment", 8) == 0) {
				/* Comment */
				for (; str[i] != '\n'; i++)
					buf = g_string_append_c(buf, str[i]);

				/* Dont forget the newline character */
				buf = g_string_append_c(buf, str[i]);

				obj = G_OBJECT(xfbib_bibtex_comment_new(G_OBJECT(elements), buf->str));
#if G_SEQUENCE
				g_sequence_append(elements->seq, obj);
#else
				elements->list = g_list_append(elements->list, obj);
#endif
				buf = g_string_erase(buf, 0, -1);
			} else if (g_ascii_strncasecmp(&str[i], "@preamble", 9) == 0) {
				/* Preamble */
				if (parse_balanced_brackets(buf, str, &i)) {
					obj = G_OBJECT(xfbib_bibtex_preamble_new(G_OBJECT(elements)));
					if (xfbib_bibtex_preamble_parse(XFBIB_BIBTEX_PREAMBLE(obj), buf->str)) {
#if G_SEQUENCE
						g_sequence_append(elements->seq, obj);
#else
						elements->list = g_list_append(elements->list, obj);
#endif
					}
				}
				buf = g_string_erase(buf, 0, -1);
			} else if (g_ascii_strncasecmp(&str[i], "@string", 7) == 0) {
				/* String */
				if (parse_balanced_brackets(buf, str, &i)) {
					obj = G_OBJECT(xfbib_bibtex_string_new(G_OBJECT(elements)));
					if (xfbib_bibtex_string_parse(XFBIB_BIBTEX_STRING(obj), buf->str)) {
						elements->n_strings++;
#if G_SEQUENCE
						g_sequence_append(elements->seq, obj);
#else
						elements->list = g_list_append(elements->list, obj);
#endif
					}
				}
				buf = g_string_erase(buf, 0, -1);
			} else if (g_regex_match (regex, &str[i], 0, NULL)) {
				/* Entry */
				if (parse_balanced_brackets(buf, str, &i)) {
					obj = G_OBJECT(xfbib_bibtex_entry_new(G_OBJECT(elements)));
					if (xfbib_bibtex_entry_parse(XFBIB_BIBTEX_ENTRY(obj), buf->str)) {
						elements->n_entries++;
#if G_SEQUENCE
						g_sequence_append(elements->seq, obj);
#else
						elements->list = g_list_append(elements->list, obj);
#endif
					}
				}
				buf = g_string_erase(buf, 0, -1);
			} else {
				/* Error */
				g_warning(_("Unknown entry type\n"));
				while (str[++i] != '@')
					continue;
			}
		} else {
			buf = g_string_append_c(buf, str[i]);
		}
	}

	g_string_free(buf, TRUE);
	g_regex_unref(regex);
	return TRUE;
}

void
xfbib_bibtex_foreach(XfbibBibtex *elements, GFunc func, gpointer data)
{
	g_return_if_fail(XFBIB_IS_BIBTEX(elements) && func != NULL);
#if G_SEQUENCE
	g_sequence_foreach(elements->seq, func, data);
#else
	g_list_foreach(elements->list, func, data);
#endif
}

gint
xfbib_bibtex_get_n(XfbibBibtex *elements, enum XFBIB_BIBTEX_TYPE type)
{
	g_return_val_if_fail(XFBIB_IS_BIBTEX(elements), 0);
	if (type == TYPE_ENTRY)
		return elements->n_entries;
	else if (type == TYPE_STRING)
		return elements->n_strings;
	return 0;
}

GObject *
xfbib_bibtex_get_nth(XfbibBibtex *elements, gint n, enum XFBIB_BIBTEX_TYPE type)
{
	GObject *tmp;
	gint found;
	g_return_val_if_fail(XFBIB_IS_BIBTEX(elements) && 
			((type == TYPE_ENTRY && n < elements->n_entries) ||
			(type == TYPE_STRING && n < elements->n_strings)), NULL);

#if G_SEQUENCE
	GSequenceIter *iter;
	for (iter = g_sequence_get_begin_iter(elements->seq), found = -1;
			!g_sequence_iter_is_end(iter); iter = g_sequence_iter_next(iter)) {
		tmp = g_sequence_get(iter);
		if (type == TYPE_ENTRY && XFBIB_IS_BIBTEX_ENTRY(tmp)) {
			if (++found == n) {
				return tmp;
			}
		} else if (type == TYPE_STRING && XFBIB_IS_BIBTEX_STRING(tmp)) {
			if (++found == n) {
				return tmp;
			}
		}
	}
#else
	gint i, len;
	len = g_list_length(elements->list);
	for (i = 0, found = -1; i < len; i++) {
		tmp = g_list_nth_data(elements->list, i);
		if (type == TYPE_ENTRY && XFBIB_IS_BIBTEX_ENTRY(tmp)) {
			if (++found == n) {
				return tmp;
			}
		} else if (type == TYPE_STRING && XFBIB_IS_BIBTEX_STRING(tmp)) {
			if (++found == n) {
				return tmp;
			}
		}
	}
#endif
	return NULL;
}


gint
xfbib_bibtex_get_position(XfbibBibtex *elements, GObject *obj)
{
	GObject *tmp;
	gint n;
	g_return_val_if_fail(XFBIB_IS_BIBTEX(elements), -1);

#if G_SEQUENCE
	GSequenceIter *iter;
	for (iter = g_sequence_get_begin_iter(elements->seq), n = -1;
			!g_sequence_iter_is_end(iter); iter = g_sequence_iter_next(iter)) {
		tmp = g_sequence_get(iter);
		if (XFBIB_IS_BIBTEX_ENTRY(obj) && XFBIB_IS_BIBTEX_ENTRY(tmp)) {
			n++;
			if (G_OBJECT(tmp) == obj) {
				return n;
			}
		} else if (XFBIB_IS_BIBTEX_STRING(obj) && XFBIB_IS_BIBTEX_STRING(tmp)) {
			n++;
			if (G_OBJECT(tmp) == obj) {
				return n;
			}
		}
	}
#else
	gint i, len;
	len = g_list_length(elements->list);
	for (i = 0, n = -1; i < len; i++) {
		tmp = g_list_nth_data(elements->list, i);
		if (XFBIB_IS_BIBTEX_ENTRY(obj) && XFBIB_IS_BIBTEX_ENTRY(tmp)) {
			n++;
			if (G_OBJECT(tmp) == obj) {
				return n;
			}
		} else if (XFBIB_IS_BIBTEX_STRING(obj) && XFBIB_IS_BIBTEX_STRING(tmp)) {
			n++;
			if (G_OBJECT(tmp) == obj) {
				return n;
			}
		}
	}
#endif
	return -1;
}

GObject *
xfbib_bibtex_get_next(XfbibBibtex *elements, GObject *obj)
{
	GObject *tmp;
	g_return_val_if_fail(XFBIB_IS_BIBTEX(elements), NULL);

#if G_SEQUENCE
	GSequenceIter *iter;
	for (iter = g_sequence_get_begin_iter(elements->seq);
			!g_sequence_iter_is_end(iter); iter = g_sequence_iter_next(iter)) {
		tmp = g_sequence_get(iter);
		if (tmp == obj) {
			for (iter = g_sequence_iter_next(iter); !g_sequence_iter_is_end(iter);
					iter = g_sequence_iter_next(iter)) {
				tmp = g_sequence_get(iter);
				if (XFBIB_IS_BIBTEX_ENTRY(obj) && XFBIB_IS_BIBTEX_ENTRY(tmp)) {
					return tmp;
				} else if (XFBIB_IS_BIBTEX_STRING(obj) && XFBIB_IS_BIBTEX_STRING(tmp)) {
					return tmp;
				}
			}
		}
	}
#else
	gint i, len;
	len = g_list_length(elements->list);
	for (i = 0; i < len; i++) {
		tmp = g_list_nth_data(elements->list, i);
		if (tmp == obj) {
			for (i++; i < len; i++) {
				tmp = g_list_nth_data(elements->list, i);
				if (XFBIB_IS_BIBTEX_ENTRY(obj) && XFBIB_IS_BIBTEX_ENTRY(tmp)) {
					return tmp;
				} else if (XFBIB_IS_BIBTEX_STRING(obj) && XFBIB_IS_BIBTEX_STRING(tmp)) {
					return tmp;
				}
			}
		}
	}
#endif
	return NULL;
}


void
xfbib_bibtex_remove(XfbibBibtex *elements, GObject *data)
{
	GObject *obj;
	g_return_if_fail(XFBIB_IS_BIBTEX(elements));
#if G_SEQUENCE
	GSequenceIter *iter;
	for (iter = g_sequence_get_begin_iter(elements->seq);
			!g_sequence_iter_is_end(iter); iter = g_sequence_iter_next(iter)) {
		if ((obj = g_sequence_get(iter)) == data) {
			if(XFBIB_IS_BIBTEX_ENTRY(obj))
				elements->n_entries--;
			else if(XFBIB_IS_BIBTEX_STRING(obj))
				elements->n_strings--;
			g_sequence_remove(iter);
		}
	}
#else
	gint i, len;
	len = g_list_length(elements->list);
	for (i = 0; i < len; i++) {
		if ((obj = g_list_nth_data(elements->list, i)) == data) {
			if(XFBIB_IS_BIBTEX_ENTRY(obj))
				elements->n_entries--;
			else if(XFBIB_IS_BIBTEX_STRINGS(obj))
				elements->n_strings--;
			elements->list = g_list_remove(elements->list, obj);
			g_object_unref(obj);
		}
	}
#endif
}

void
xfbib_bibtex_remove_all(XfbibBibtex *elements)
{
	g_return_if_fail(XFBIB_IS_BIBTEX(elements));
#if G_SEQUENCE
	g_sequence_remove_range(g_sequence_get_begin_iter(elements->seq),
			g_sequence_get_end_iter(elements->seq));
	elements->seq = g_sequence_new(g_object_unref);
#else
	g_list_foreach (elements->list, (GFunc) g_object_unref, NULL);
	g_list_free (elements->list);
	elements->list = NULL;
#endif
	elements->n_entries = 0;
	elements->n_strings = 0;
}

void
xfbib_bibtex_add(XfbibBibtex *elements, GObject *obj)
{
	g_return_if_fail(XFBIB_IS_BIBTEX(elements));
	g_return_if_fail(XFBIB_IS_BIBTEX_ENTRY(obj) || XFBIB_IS_BIBTEX_STRING(obj));

#if G_SEQUENCE
	g_sequence_append(elements->seq, G_OBJECT(obj));
#else
	elements->list = g_list_append(elements->list, G_OBJECT(obj));
#endif
	if (XFBIB_IS_BIBTEX_ENTRY(obj))
		elements->n_entries++;
	else if (XFBIB_IS_BIBTEX_STRING(obj))
		elements->n_strings++;
}

/*
 * Return the position of the new object, counting from 0
 */
gint
xfbib_bibtex_replace(XfbibBibtex *elements, GObject *old_obj, GObject *new_obj)
{
	GObject *obj;
	gint n = 0;
	g_return_val_if_fail(XFBIB_IS_BIBTEX(elements) &&
			((XFBIB_IS_BIBTEX_ENTRY(old_obj) && 
			XFBIB_IS_BIBTEX_ENTRY(new_obj)) ||
			(XFBIB_IS_BIBTEX_STRING(old_obj) && 
			XFBIB_IS_BIBTEX_STRING(new_obj))), -1);

#if G_SEQUENCE
	GSequenceIter *iter;
	for (iter = g_sequence_get_begin_iter(elements->seq);
			!g_sequence_iter_is_end(iter); iter = g_sequence_iter_next(iter)) {
		obj = g_sequence_get(iter);
		if ((XFBIB_IS_BIBTEX_ENTRY(old_obj) && XFBIB_IS_BIBTEX_ENTRY(obj)) ||
			(XFBIB_IS_BIBTEX_STRING(old_obj) && XFBIB_IS_BIBTEX_STRING(obj))) {
			n++;
		}

		if (obj == old_obj) {
			g_sequence_insert_before(iter, new_obj);
			g_sequence_remove(iter);
			return n;
		}
	}
#else
	gint i, len;
	len = g_list_length(elements->list);
	for (i = 0; i < len; i++) {
		obj = g_list_nth(elements->list, i);
		if (XFBIB_IS_BIBTEX_ENTRY(old_obj) && XFBIB_IS_BIBTEX_ENTRY(obj))
			n++;
		else if (XFBIB_IS_BIBTEX_STRING(old_obj) && XFBIB_IS_BIBTEX_STRING(obj))		
			n++;

		if (obj->data == old_obj) {
			elements->list = g_list_insert_before(elements->list, obj, new_obj);
			elements->list = g_list_remove(elements->list, obj);
			return n;
		}
	}
#endif
	return -1;
}

GObject *
xfbib_bibtex_find(XfbibBibtex *elements, GCompareFunc func, gconstpointer ptr)
{
	GObject *tmp;
	g_return_val_if_fail(XFBIB_IS_BIBTEX(elements), NULL);

#if G_SEQUENCE
	GSequenceIter *iter;
	for (iter = g_sequence_get_begin_iter(elements->seq);
			!g_sequence_iter_is_end(iter); iter = g_sequence_iter_next(iter)) {
		tmp = g_sequence_get(iter);
#else
	gint i, len;
	len = g_list_length(elements->list);
	for (i = 0, n = -1; i < len; i++) {
		tmp = g_list_nth_data(elements->list, i);
#endif
		if (!func(ptr, tmp)) {
			return tmp;
		}
	}
	return NULL;
}

