/*
 * Copyright (c) 2008 Jesper Karlsson & David Gustafsson
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif
#include <string.h>
#include <gtk/gtk.h>
#include <libxfce4ui/libxfce4ui.h>

#include "xfbib-list-store.h"
#include "xfbib-state.h"
#include "xfbib-bibtex.h"
#include "xfbib-bibtex-comment.h"
#include "xfbib-bibtex-entry.h"
#include "xfbib-bibtex-field.h"
#include "xfbib-bibtex-preamble.h"
#include "xfbib-bibtex-string.h"


struct _XfbibListStore
{
	GtkListStore parent;
	XfbibState *state;
	gint stamp;
	enum XFBIB_BIBTEX_TYPE type;
};

typedef struct _XfbibListStoreClass
{
	GtkListStoreClass parent;
} XfbibListStoreClass;


static void xfbib_list_store_class_init(XfbibListStoreClass *klass);
static void xfbib_list_store_tree_model_init (GtkTreeModelIface *iface);

static void xfbib_list_store_init(XfbibListStore *instance);
static void xfbib_list_store_finalize(GObject *obj);

static GtkTreeModelFlags xfbib_list_store_get_flags (GtkTreeModel *);
static gint xfbib_list_store_get_n_columns (GtkTreeModel *);
static GType xfbib_list_store_get_column_type (GtkTreeModel *, gint);
static gboolean xfbib_list_store_get_iter (GtkTreeModel *, GtkTreeIter *, GtkTreePath *);
static GtkTreePath *xfbib_list_store_get_path (GtkTreeModel *, GtkTreeIter *);
static void xfbib_list_store_get_value (GtkTreeModel *, GtkTreeIter *, gint , GValue *);
static gboolean xfbib_list_store_iter_next (GtkTreeModel *, GtkTreeIter *);
static gboolean xfbib_list_store_iter_children (GtkTreeModel *, GtkTreeIter *, GtkTreeIter *);
static gboolean xfbib_list_store_iter_has_child (GtkTreeModel *, GtkTreeIter *);
static gint xfbib_list_store_iter_n_children (GtkTreeModel *, GtkTreeIter *);
static gboolean xfbib_list_store_iter_nth_child (GtkTreeModel *, GtkTreeIter *, GtkTreeIter *, gint);
static gboolean xfbib_list_store_iter_parent (GtkTreeModel *, GtkTreeIter *, GtkTreeIter *);


static GObjectClass *xfbib_list_store_parent_class = NULL;

GType
xfbib_list_store_get_type (void)
{
	static GType type = 0;
	if (type == 0) {
		static const GTypeInfo info = {
			sizeof (XfbibListStoreClass),
			NULL,   /* base_init */
			NULL,   /* base_finalize */
			(GClassInitFunc) xfbib_list_store_class_init,   /* class_init */
			NULL,   /* class_finalize */
			NULL,   /* class_data */
			sizeof (XfbibListStore),
			0,      /* n_preallocs */
			(GInstanceInitFunc) xfbib_list_store_init    /* instance_init */
		};
		static const GInterfaceInfo tree_model_info =
		{
			(GInterfaceInitFunc) xfbib_list_store_tree_model_init,
			NULL,
			NULL
		};

		type = g_type_register_static (G_TYPE_OBJECT, "XfbibListStore", &info, 0);

		/* Register the GtkListStore interface */
		g_type_add_interface_static (type, GTK_TYPE_TREE_MODEL, &tree_model_info);
	}
	return type;
}

static void
xfbib_list_store_class_init(XfbibListStoreClass *klass)
{
	GObjectClass *object_class = (GObjectClass *)klass;

	xfbib_list_store_parent_class = g_type_class_peek_parent (klass);

	object_class->finalize = xfbib_list_store_finalize;
}

static void
xfbib_list_store_tree_model_init (GtkTreeModelIface *iface)
{
	iface->get_flags = xfbib_list_store_get_flags;
	iface->get_n_columns = xfbib_list_store_get_n_columns;
	iface->get_column_type = xfbib_list_store_get_column_type;
	iface->get_iter = xfbib_list_store_get_iter;
	iface->get_path = xfbib_list_store_get_path;
	iface->get_value = xfbib_list_store_get_value;
	iface->iter_next = xfbib_list_store_iter_next;
	iface->iter_children = xfbib_list_store_iter_children;
	iface->iter_has_child = xfbib_list_store_iter_has_child;
	iface->iter_n_children = xfbib_list_store_iter_n_children;
	iface->iter_nth_child = xfbib_list_store_iter_nth_child;
	iface->iter_parent = xfbib_list_store_iter_parent;
}


static void
xfbib_list_store_init(XfbibListStore *instance)
{
	instance->stamp = g_random_int();
	instance->state = xfbib_state_new();
}

static void
xfbib_list_store_finalize(GObject *obj)
{
	G_OBJECT_CLASS(xfbib_list_store_parent_class)->finalize(obj);
}


static GtkTreeModelFlags
xfbib_list_store_get_flags (GtkTreeModel *tree_model)
{
	return GTK_TREE_MODEL_ITERS_PERSIST | GTK_TREE_MODEL_LIST_ONLY;
}

static gint
xfbib_list_store_get_n_columns (GtkTreeModel *tree_model)
{
	XfbibListStore *list_store;
	g_return_val_if_fail(XFBIB_IS_LIST_STORE(tree_model), 0);

	list_store = XFBIB_LIST_STORE(tree_model);

	switch (list_store->type) {
		case TYPE_ENTRY:
			return XFBIB_BIBTEX_FIELD_N_FIELDS;
		case TYPE_STRING:
			return XFBIB_BIBTEX_STRING_N_STRINGS;
		default:
			return 0;
	}
}

static GType
xfbib_list_store_get_column_type (GtkTreeModel *tree_model, gint index)
{
	XfbibListStore *list_store;
	g_return_val_if_fail((XFBIB_IS_LIST_STORE(tree_model) || index >= 0), G_TYPE_INVALID);

	list_store = XFBIB_LIST_STORE(tree_model);
	switch (list_store->type) {
		case TYPE_ENTRY:
			if (index >= XFBIB_BIBTEX_FIELD_N_FIELDS)
				return G_TYPE_INVALID;
			return G_TYPE_STRING;
		case TYPE_STRING:
			if (index >= XFBIB_BIBTEX_STRING_N_STRINGS)
				return G_TYPE_INVALID;
			return G_TYPE_STRING;
		default:
			return G_TYPE_INVALID;
	}
}

static gboolean
xfbib_list_store_get_iter (GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreePath *path)
{
	XfbibListStore *list_store;
	XfbibBibtex *elements;
	gint n;

	g_return_val_if_fail(XFBIB_IS_LIST_STORE(tree_model), FALSE);

	list_store = XFBIB_LIST_STORE(tree_model);
	elements = xfbib_state_get_bibtex_elements(list_store->state);

	n = gtk_tree_path_get_indices(path)[0];

	switch (list_store->type) {
		case TYPE_ENTRY:
		case TYPE_STRING:
			if (n >= xfbib_bibtex_get_n(elements, list_store->type))
				return FALSE;

			iter->stamp = list_store->stamp;
			iter->user_data = xfbib_bibtex_get_nth(elements, n, list_store->type);
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

static GtkTreePath *
xfbib_list_store_get_path (GtkTreeModel *tree_model, GtkTreeIter *iter)
{
	XfbibListStore *list_store;
	XfbibBibtex *elements;
	GtkTreePath *path;
	gint n;

	g_return_val_if_fail(XFBIB_IS_LIST_STORE(tree_model), NULL);

	list_store = XFBIB_LIST_STORE(tree_model);
	elements = xfbib_state_get_bibtex_elements(list_store->state);

	if (list_store->stamp != iter->stamp)
		return NULL;

	switch (list_store->type) {
		case TYPE_ENTRY:
		case TYPE_STRING:
			if ((n = xfbib_bibtex_get_position(elements,
							iter->user_data)) == -1) {
				return NULL;
			}
			break;
		default:
			return NULL;
	}

	path = gtk_tree_path_new();
	gtk_tree_path_append_index(path, n);

	return path;
}

struct field_properties xfbib_bibtex_field_constants_list[] = {
	{"Address", ("Address of publisher")},
	{"Annote", ("Annotation for annotated bibliography styles")},
	{"Author", ("Name(s) of the author(s), separated by 'and' if more than one")},
	{"Booktitle", ("Title of the book, if only part of it is being cited")},
	{"Chapter", ("Chapter number")},
	{"Crossref", ("Citation key of the cross-referenced entry")},
	{"Edition", ("Edition of the book (such as \"first\" or \"second\")")},
	{"Editor", ("Name(s) of the editor(s), separated by 'and' if more than one")},
	{"E-print", ("Specification of electronic publication")},
	{"HowPublished", ("Publishing method if the method is nonstandard")},
	{"Institution", ("Institution that was involved in the publishing")},
	{"Journal", ("Journal or magazine in which the work was published")},
	{"Key", ("Hidden field used for specifying or overriding the alphabetical order of entries")},
	{"Month", ("Month of publication or creation if unpublished")},
	{"Note", ("Miscellaneous extra information")},
	{"Number", ("Number of journal, magazine, or tech-report")},
	{"Organization", ("Sponsor of the conference")},
	{"Pages", ("Page numbers separated by commas or double-hyphens")},
	{"Publisher", ("Name of publisher")},
	{"School", ("School where thesis was written")},
	{"Series", ("Series of books in which the book was published")},
	{"Title", ("Title of the work")},
	{"Type", ("Type of technical report")},
	{"URL", ("Internet address")},
	{"Volume", ("Number of the volume")},
	{"Year", ("Year of publication or creation if unpublished")}
};

static void
xfbib_list_store_get_value (GtkTreeModel *tree_model, GtkTreeIter *iter, gint column, GValue *value)
{
	XfbibListStore *list_store;
	XfbibBibtexField *field;
	GList *list;
	gchar *tmp;
	gint i;

	g_return_if_fail(XFBIB_IS_LIST_STORE(tree_model));

	list_store = XFBIB_LIST_STORE(tree_model);

	if (list_store->stamp != iter->stamp)
		return;

	/* Just to make the compiler stop nagging */
	field = NULL;

	g_value_init (value, G_TYPE_STRING);

	switch (list_store->type) {
		case TYPE_ENTRY:
			g_return_if_fail(column < XFBIB_BIBTEX_FIELD_N_FIELDS
					 && XFBIB_IS_BIBTEX_ENTRY(iter->user_data));


			list = xfbib_bibtex_entry_get_fields(XFBIB_BIBTEX_ENTRY(iter->user_data));

			for (i = 0; i < g_list_length(list); i++) {
				field = g_list_nth_data(list, i);
				if (strcasecmp(xfbib_bibtex_field_get_variable(field),
							xfbib_bibtex_field_constants_list[column].label) == 0) {
					break;
				}
			}
			
			if (i >= g_list_length(list)) {
				/* Did not find a matching field */
				return;
			}

			tmp = xfbib_bibtex_value_get_str(xfbib_bibtex_field_get_value(field));
			g_value_set_string(value, (tmp == NULL) ? "" : tmp);
			break;
		case TYPE_STRING:
			g_return_if_fail(column < XFBIB_BIBTEX_STRING_N_STRINGS &&
					XFBIB_IS_BIBTEX_STRING(iter->user_data));

			if (column == XFBIB_BIBTEX_STRING_VARIABLE) {
				g_value_set_string(value, xfbib_bibtex_string_get_variable(iter->user_data));
			} else if (column == XFBIB_BIBTEX_STRING_VALUE) {
				tmp = xfbib_bibtex_value_get_str(xfbib_bibtex_string_get_value(iter->user_data));
				g_value_set_string(value, (tmp == NULL) ? "" : tmp);
			}

			break;
		default:
			return;
	}
}

static gboolean
xfbib_list_store_iter_next (GtkTreeModel *tree_model, GtkTreeIter *iter)
{
	XfbibListStore *list_store;
	XfbibBibtex *elements;
	GObject *next;

	g_return_val_if_fail(XFBIB_IS_LIST_STORE(tree_model), FALSE);

	list_store = XFBIB_LIST_STORE(tree_model);
	elements = xfbib_state_get_bibtex_elements(list_store->state);

	g_return_val_if_fail(list_store->stamp == iter->stamp, FALSE);

	if (iter == NULL || iter->user_data == NULL)
		return FALSE;

	switch (list_store->type) {
		case TYPE_ENTRY:
		case TYPE_STRING:
			if ((next = xfbib_bibtex_get_next(elements, iter->user_data)) == NULL) {
				iter->stamp = 0;
				return FALSE;
			}
			iter->stamp = list_store->stamp;
			iter->user_data = next;
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

static gboolean
xfbib_list_store_iter_children (GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent)
{
	XfbibListStore *list_store;
	XfbibBibtex *elements;

	g_return_val_if_fail(XFBIB_IS_LIST_STORE(tree_model), FALSE);

	list_store = XFBIB_LIST_STORE(tree_model);
	elements = xfbib_state_get_bibtex_elements(list_store->state);

	/* This is only a list and therefore nodes have have no children */
	if (parent != NULL)
		return FALSE;

	switch (list_store->type) {
		case TYPE_ENTRY:
		case TYPE_STRING:
			if (xfbib_bibtex_get_n(elements, list_store->type) == 0)
				return FALSE;

			iter->stamp = list_store->stamp;
			if ((iter->user_data = xfbib_bibtex_get_nth(elements, 0, list_store->type)) == NULL)
				return FALSE;
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

static gboolean
xfbib_list_store_iter_has_child (GtkTreeModel *tree_model, GtkTreeIter *iter)
{
	return FALSE;
}

static gint
xfbib_list_store_iter_n_children (GtkTreeModel *tree_model, GtkTreeIter *iter)
{
	XfbibListStore *list_store;
	XfbibBibtex *elements;

	g_return_val_if_fail(XFBIB_IS_LIST_STORE(tree_model), -1);

	list_store = XFBIB_LIST_STORE(tree_model);
	elements = xfbib_state_get_bibtex_elements(list_store->state);

	switch (list_store->type) {
		case TYPE_ENTRY:
		case TYPE_STRING:
			if (iter == NULL)
				return xfbib_bibtex_get_n(elements, list_store->type);
			break;
		default:
			return -1;
	}
	return 0;
}

static gboolean
xfbib_list_store_iter_nth_child (GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *parent, gint n)
{
	XfbibListStore *list_store;
	XfbibBibtex *elements;

	g_return_val_if_fail(XFBIB_IS_LIST_STORE(tree_model), FALSE);

	list_store = XFBIB_LIST_STORE(tree_model);
	elements = xfbib_state_get_bibtex_elements(list_store->state);

	/* This is only a list and therefore nodes have have no children */
	if (parent != NULL)
		return FALSE;

	switch (list_store->type) {
		case TYPE_ENTRY:
		case TYPE_STRING:
			if (n >= xfbib_bibtex_get_n(elements, list_store->type))
				return FALSE;

			iter->stamp = list_store->stamp;
			if ((iter->user_data = xfbib_bibtex_get_nth(elements, n, list_store->type)) == NULL)
				return FALSE;
			break;
		default:
			return FALSE;
	}
	return TRUE;
}

static gboolean
xfbib_list_store_iter_parent (GtkTreeModel *tree_model, GtkTreeIter *iter, GtkTreeIter *child)
{
	return FALSE;
}

XfbibListStore *
xfbib_list_store_new(enum XFBIB_BIBTEX_TYPE type)
{
	XfbibListStore *list_store;
	list_store = g_object_new(XFBIB_TYPE_LIST_STORE, NULL);
	list_store->type = type;
	return list_store;
}


gboolean
xfbib_list_store_remove(XfbibListStore *list_store, GtkTreeIter *iter)
{
	XfbibBibtex *elements;
	GtkTreePath *path;
	GtkTreeIter next;
	gboolean more;

	g_return_val_if_fail(XFBIB_IS_LIST_STORE (list_store), FALSE);
	g_return_val_if_fail(list_store->stamp == iter->stamp, FALSE);
	g_return_val_if_fail(iter->user_data != NULL, FALSE);

	path = xfbib_list_store_get_path(GTK_TREE_MODEL(list_store), iter);

	next.stamp = iter->stamp;
	next.user_data = iter->user_data;
	more = xfbib_list_store_iter_next(GTK_TREE_MODEL(list_store), &next);

	elements = xfbib_state_get_bibtex_elements(list_store->state);
	xfbib_bibtex_remove(elements, iter->user_data);

	gtk_tree_model_row_deleted (GTK_TREE_MODEL (list_store), path);
	gtk_tree_path_free (path);

	if (more == FALSE) {
		iter->stamp = 0;
		return FALSE;
	}

	iter->stamp = list_store->stamp;
	iter->user_data = next.user_data;
	return TRUE;
}


void
xfbib_list_store_clear(XfbibListStore *list_store)
{
	XfbibBibtex *elements;
	g_return_if_fail(XFBIB_IS_LIST_STORE(list_store));

	elements = xfbib_state_get_bibtex_elements(list_store->state);
	xfbib_bibtex_remove_all(elements);

	/* Old iters are not valid any more so change the stamp */
	list_store->stamp = g_random_int();
}


void
xfbib_list_store_row_inserted(XfbibListStore *list_store, gint pos)
{
	GtkTreePath *path;
	GtkTreeIter iter;
	g_return_if_fail(XFBIB_IS_LIST_STORE(list_store));

	path = gtk_tree_path_new();
	gtk_tree_path_append_index(path, pos);

	gtk_tree_model_get_iter(GTK_TREE_MODEL(list_store), &iter, path);

	gtk_tree_model_row_inserted(GTK_TREE_MODEL(list_store), path, &iter);

	gtk_tree_path_free(path);
}

void
xfbib_list_store_row_changed(XfbibListStore *list_store, gint pos)
{
	GtkTreePath *path;
	GtkTreeIter iter;
	g_return_if_fail(XFBIB_IS_LIST_STORE(list_store));

	path = gtk_tree_path_new();
	gtk_tree_path_append_index(path, pos);

	gtk_tree_model_get_iter(GTK_TREE_MODEL(list_store), &iter, path);

	gtk_tree_model_row_changed(GTK_TREE_MODEL(list_store), path, &iter);

	gtk_tree_path_free(path);
}

GObject *
xfbib_list_store_get(XfbibListStore *list_store, GtkTreeIter *iter)
{
	g_return_val_if_fail(XFBIB_IS_LIST_STORE (list_store), NULL);
	g_return_val_if_fail(list_store->stamp == iter->stamp, NULL);
	g_return_val_if_fail(iter->user_data != NULL, NULL);

	return G_OBJECT(iter->user_data);
}

void
xfbib_list_store_add(XfbibListStore *list_store, GObject *obj)
{
	XfbibBibtex *elements;
	gint n;
	g_return_if_fail(XFBIB_IS_LIST_STORE(list_store) &&
			(XFBIB_BIBTEX_ENTRY(obj) ||
			 XFBIB_BIBTEX_STRING(obj)));

	elements = xfbib_state_get_bibtex_elements(list_store->state);

	xfbib_bibtex_add(elements, obj);

	n = xfbib_bibtex_get_n(elements, list_store->type);
	xfbib_list_store_row_inserted(list_store, n-1);
}

void
xfbib_list_store_replace(XfbibListStore *list_store, GObject *old_obj, GObject *new_obj)
{
	XfbibBibtex *elements;
	gint n;
	
	g_return_if_fail(XFBIB_IS_LIST_STORE(list_store) &&
			XFBIB_BIBTEX_ENTRY(old_obj) &&
			XFBIB_BIBTEX_ENTRY(new_obj));
	
	elements = xfbib_state_get_bibtex_elements(list_store->state);

	n = xfbib_bibtex_replace(elements, old_obj, new_obj);
	
	xfbib_list_store_row_changed(list_store, n);
}
